# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from urllib.parse import urlencode
from datetime import datetime
from inspect import isawaitable
from typing import Any, Dict, Union

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.test import RequestFactory, TestCase

from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.django import middleware
from supertokens_python.framework.django.django_response import (
    DjangoResponse as SuperTokensDjangoWrapper,
)
from supertokens_python.recipe import emailpassword, session, thirdparty
from supertokens_python.recipe.emailpassword.interfaces import APIInterface, APIOptions
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from supertokens_python.recipe.session.framework.django.asyncio import verify_session
from tests.utils import clean_st, reset, setup_st, start_st


def get_cookies(response: HttpResponse) -> Dict[str, Any]:
    cookies: Dict[str, Any] = {}
    for key, morsel in response.cookies.items():
        cookies[key] = {"value": morsel.value, "name": key}
        for k, v in morsel.items():
            if (k in ("secure", "httponly")) and v == "":
                cookies[key][k] = None
            elif k == "samesite":
                if len(v) > 0 and v[-1] == ",":
                    v = v[:-1]
                cookies[key][k] = v
            else:
                cookies[key][k] = v
    return cookies


async def create_new_session_view(request: HttpRequest):
    await create_new_session(request, "user_id")
    return JsonResponse({"foo": "bar"})


async def refresh_view(request: HttpRequest):
    await refresh_session(request)
    return JsonResponse({"foo": "bar"})


async def custom_response_view(_: HttpRequest):
    pass


async def logout_view(request: HttpRequest):
    session: Union[None, SessionContainer] = await get_session(request, True)
    if session is None:
        raise Exception("Should never come here")
    await session.revoke_session()
    return JsonResponse({"foo": "bar"})


async def handle_view(request: HttpRequest):
    session: Union[None, SessionContainer] = await get_session(request, True)
    if session is None:
        raise Exception("Should never come here")
    return JsonResponse({"s": session.get_handle()})


@verify_session(session_required=False)
async def optional_session(request: HttpRequest):
    session: Union[None, SessionContainer] = request.supertokens  # type: ignore
    if session is None:
        return JsonResponse({"s": "empty session"})
    return JsonResponse({"s": session.get_handle()})


class SupertokensTest(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        reset()
        clean_st()
        setup_st()

    def tearDown(self):
        reset()
        clean_st()

    async def test_login_refresh(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                session.init(
                    anti_csrf="VIA_TOKEN",
                    get_token_transfer_method=lambda _, __, ___: "cookie",
                    cookie_domain="supertokens.io",
                )
            ],
        )

        start_st()

        my_middleware = middleware(create_new_session_view)
        request = self.factory.get("/login", {"user_id": "user_id"})
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp

        my_middleware = middleware(refresh_view)
        request = self.factory.get("/refresh", {"user_id": "user_id"})
        cookies = get_cookies(response)

        assert len(cookies["sAccessToken"]["value"]) > 0
        assert len(cookies["sRefreshToken"]["value"]) > 0

        request.COOKIES["sRefreshToken"] = cookies["sRefreshToken"]["value"]
        request.META["HTTP_ANTI_CSRF"] = response.headers["anti-csrf"]
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        refreshed_cookies = get_cookies(response)

        assert (
            refreshed_cookies["sAccessToken"]["value"]
            != cookies["sAccessToken"]["value"]
        )
        assert (
            refreshed_cookies["sRefreshToken"]["value"]
            != cookies["sRefreshToken"]["value"]
        )
        assert response.headers["anti-csrf"] is not None
        assert (
            refreshed_cookies["sAccessToken"]["domain"]
            == cookies["sAccessToken"]["domain"]
        )
        assert (
            refreshed_cookies["sRefreshToken"]["domain"]
            == cookies["sRefreshToken"]["domain"]
        )
        assert (
            refreshed_cookies["sAccessToken"]["secure"]
            == cookies["sAccessToken"]["secure"]
        )
        assert (
            refreshed_cookies["sRefreshToken"]["secure"]
            == cookies["sRefreshToken"]["secure"]
        )

    async def test_login_logout(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                session.init(
                    anti_csrf="VIA_TOKEN",
                    get_token_transfer_method=lambda _, __, ___: "cookie",
                    cookie_domain="supertokens.io",
                )
            ],
        )

        start_st()

        my_middleware = middleware(create_new_session_view)
        request = self.factory.get("/login", {"user_id": "user_id"})
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        cookies = get_cookies(response)

        assert len(cookies["sAccessToken"]["value"]) > 0
        assert len(cookies["sRefreshToken"]["value"]) > 0

        my_middleware = middleware(logout_view)
        request = self.factory.post("/logout", {"user_id": "user_id"})

        request.COOKIES["sAccessToken"] = cookies["sAccessToken"]["value"]
        request.META["HTTP_ANTI_CSRF"] = response.headers["anti-csrf"]
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        logout_cookies = get_cookies(response)
        assert response.headers.get("anti-csrf") is None  # type: ignore
        assert logout_cookies["sAccessToken"]["value"] == ""
        assert logout_cookies["sRefreshToken"]["value"] == ""

    async def test_login_handle(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                session.init(
                    anti_csrf="VIA_TOKEN",
                    cookie_domain="supertokens.io",
                    get_token_transfer_method=lambda _, __, ___: "cookie",
                )
            ],
        )

        start_st()

        my_middleware = middleware(create_new_session_view)
        request = self.factory.get("/login", {"user_id": "user_id"})
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        cookies = get_cookies(response)

        assert len(cookies["sAccessToken"]["value"]) > 0
        assert len(cookies["sRefreshToken"]["value"]) > 0

        try:
            datetime.strptime(
                cookies["sAccessToken"]["expires"], "%a, %d %b %Y %H:%M:%S GMT"
            )
        except ValueError:
            assert False, "cookies expiry time doesn't have the correct format"

        my_middleware = middleware(handle_view)
        request = self.factory.get("/handle", {"user_id": "user_id"})

        request.COOKIES["sAccessToken"] = cookies["sAccessToken"]["value"]
        request.META["HTTP_ANTI_CSRF"] = response.headers["anti-csrf"]
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        assert "s" in json.loads(
            response.content
        )  # FIXME: Getting Unauthorized in body. Why?
        handle_cookies = get_cookies(response)

        assert not handle_cookies

    async def test_login_refresh_error_handler(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                session.init(
                    anti_csrf="VIA_TOKEN",
                    cookie_domain="supertokens.io",
                    get_token_transfer_method=lambda _, __, ___: "cookie",
                )
            ],
        )

        start_st()

        my_middleware = middleware(create_new_session_view)
        request = self.factory.get("/login", {"user_id": "user_id"})
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp

        my_middleware = middleware(refresh_view)
        request = self.factory.get("/refresh", {"user_id": "user_id"})
        cookies = get_cookies(response)

        assert len(cookies["sAccessToken"]["value"]) > 0
        assert len(cookies["sRefreshToken"]["value"]) > 0

        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp
        # not authorized because no access refresh token
        assert response.status_code == 401

    async def test_custom_response(self):
        def override_email_password_apis(original_implementation: APIInterface):

            original_func = original_implementation.email_exists_get

            async def email_exists_get(
                email: str, api_options: APIOptions, user_context: Dict[str, Any]
            ):
                response_dict = {"custom": True}
                api_options.response.set_status_code(203)
                api_options.response.set_json_content(response_dict)
                return await original_func(email, api_options, user_context)

            original_implementation.email_exists_get = email_exists_get
            return original_implementation

        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                emailpassword.init(
                    override=emailpassword.InputOverrideConfig(
                        apis=override_email_password_apis
                    )
                )
            ],
        )

        start_st()

        my_middleware = middleware(custom_response_view)
        request = self.factory.get("/auth/signup/email/exists?email=test@example.com")
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp

        assert response.status_code == 203
        dict_response = json.loads(response.content)
        assert dict_response["custom"]

    async def test_optional_session(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
            ],
        )

        start_st()

        my_middleware = middleware(optional_session)
        request = self.factory.get("/handle-session-optional")
        temp = my_middleware(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp

        assert response.status_code == 200
        dict_response = json.loads(response.content)
        assert dict_response["s"] == "empty session"

    async def test_thirdparty_parsing_works(self):
        init(
            supertokens_config=SupertokensConfig("http://localhost:3567"),
            app_info=InputAppInfo(
                app_name="SuperTokens Demo",
                api_domain="http://api.supertokens.io",
                website_domain="http://supertokens.io",
                api_base_path="/auth",
            ),
            framework="django",
            mode="asgi",
            recipe_list=[
                thirdparty.init(
                    sign_in_and_up_feature=thirdparty.SignInAndUpFeature(
                        providers=[
                            thirdparty.Apple(
                                client_id="4398792-io.supertokens.example.service",
                                client_key_id="7M48Y4RYDL",
                                client_team_id="YWQCXGJRJL",
                                client_private_key="-----BEGIN PRIVATE KEY-----\nMIGTAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBHkwdwIBAQQgu8gXs+XYkqXD6Ala9Sf/iJXzhbwcoG5dMh1OonpdJUmgCgYIKoZIzj0DAQehRANCAASfrvlFbFCYqn3I2zeknYXLwtH30JuOKestDbSfZYxZNMqhF/OzdZFTV0zc5u5s3eN+oCWbnvl0hM+9IW0UlkdA\n-----END PRIVATE KEY-----",
                            )
                        ]
                    )
                ),
            ],
        )

        start_st()

        data = {
            "state": "afc596274293e1587315c",
            "code": "c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA",
        }

        request = self.factory.post(
            "/auth/callback/apple",
            urlencode(data).encode(),
            content_type="application/x-www-form-urlencoded",
        )
        temp = middleware(custom_response_view)(request)
        if not isawaitable(temp):
            raise Exception("Should never come here")
        response = await temp

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.content,
            b'<html><head><script>window.location.replace("http://supertokens.io/auth/callback/apple?state=afc596274293e1587315c&code=c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA");</script></head></html>',
        )


def test_remove_header_works():
    response = HttpResponse()
    st_response = SuperTokensDjangoWrapper(response)

    st_response.set_header("foo", "bar")
    assert st_response.get_header("foo") == "bar"
    st_response.remove_header("foo")
    assert st_response.get_header("foo") is None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from supertokens_python.recipe.session.recipe_implementation import RecipeImplementation
from pytest import mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.process_state import AllowedProcessStates, ProcessState
from supertokens_python.recipe import session
from supertokens_python.recipe.session import SessionRecipe

from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@mark.asyncio
async def test_that_once_the_info_is_loaded_it_doesnt_query_again():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )
    start_st()

    session_recipe_instance = SessionRecipe.get_instance()
    if not isinstance(
        session_recipe_instance.recipe_implementation, RecipeImplementation
    ):
        raise Exception("Should never come here")
    await session_recipe_instance.recipe_implementation.get_handshake_info()

    assert (
        AllowedProcessStates.CALLING_SERVICE_IN_GET_HANDSHAKE_INFO
        in ProcessState.get_instance().history
    )

    ProcessState.get_instance().reset()
    await session_recipe_instance.recipe_implementation.get_handshake_info()

    assert (
        AllowedProcessStates.CALLING_SERVICE_IN_GET_HANDSHAKE_INFO
        not in ProcessState.get_instance().history
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union

from supertokens_python.async_to_sync_wrapper import sync
from supertokens_python.recipe.userroles.interfaces import (
    AddRoleToUserOkResult,
    CreateNewRoleOrAddPermissionsOkResult,
    DeleteRoleOkResult,
    GetAllRolesOkResult,
    GetPermissionsForRoleOkResult,
    GetRolesForUserOkResult,
    GetRolesThatHavePermissionOkResult,
    GetUsersThatHaveRoleOkResult,
    RemovePermissionsFromRoleOkResult,
    RemoveUserRoleOkResult,
    UnknownRoleError,
)


def add_role_to_user(
    user_id: str, role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[AddRoleToUserOkResult, UnknownRoleError]:
    from supertokens_python.recipe.userroles.asyncio import add_role_to_user

    return sync(add_role_to_user(user_id, role, user_context))


def remove_user_role(
    user_id: str, role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[RemoveUserRoleOkResult, UnknownRoleError]:
    from supertokens_python.recipe.userroles.asyncio import remove_user_role

    return sync(remove_user_role(user_id, role, user_context))


def get_roles_for_user(
    user_id: str, user_context: Union[Dict[str, Any], None] = None
) -> GetRolesForUserOkResult:
    from supertokens_python.recipe.userroles.asyncio import get_roles_for_user

    return sync(get_roles_for_user(user_id, user_context))


def get_users_that_have_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[GetUsersThatHaveRoleOkResult, UnknownRoleError]:
    from supertokens_python.recipe.userroles.asyncio import get_users_that_have_role

    return sync(get_users_that_have_role(role, user_context))


def create_new_role_or_add_permissions(
    role: str, permissions: List[str], user_context: Union[Dict[str, Any], None] = None
) -> CreateNewRoleOrAddPermissionsOkResult:
    from supertokens_python.recipe.userroles.asyncio import (
        create_new_role_or_add_permissions,
    )

    return sync(create_new_role_or_add_permissions(role, permissions, user_context))


def get_permissions_for_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> Union[GetPermissionsForRoleOkResult, UnknownRoleError]:
    from supertokens_python.recipe.userroles.asyncio import get_permissions_for_role

    return sync(get_permissions_for_role(role, user_context))


def remove_permissions_from_role(
    role: str, permissions: List[str], user_context: Union[Dict[str, Any], None] = None
) -> Union[RemovePermissionsFromRoleOkResult, UnknownRoleError]:
    from supertokens_python.recipe.userroles.asyncio import remove_permissions_from_role

    return sync(remove_permissions_from_role(role, permissions, user_context))


def get_roles_that_have_permission(
    permission: str, user_context: Union[Dict[str, Any], None] = None
) -> GetRolesThatHavePermissionOkResult:
    from supertokens_python.recipe.userroles.asyncio import (
        get_roles_that_have_permission,
    )

    return sync(get_roles_that_have_permission(permission, user_context))


def delete_role(
    role: str, user_context: Union[Dict[str, Any], None] = None
) -> DeleteRoleOkResult:
    from supertokens_python.recipe.userroles.asyncio import delete_role

    return sync(delete_role(role, user_context))


def get_all_roles(
    user_context: Union[Dict[str, Any], None] = None
) -> GetAllRolesOkResult:
    from supertokens_python.recipe.userroles.asyncio import get_all_roles

    return sync(get_all_roles(user_context))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict, Union

from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryInterface
from supertokens_python.recipe.emailpassword.emaildelivery.services.backward_compatibility import (
    BackwardCompatibilityService as EPBackwardCompatibilityService,
)
from supertokens_python.recipe.emailpassword.interfaces import (
    RecipeInterface as EPRecipeInterface,
)
from supertokens_python.recipe.emailpassword.utils import (
    InputResetPasswordUsingTokenFeature,
)
from supertokens_python.recipe.thirdpartyemailpassword.types import (
    EmailTemplateVars,
)
from supertokens_python.supertokens import AppInfo


class BackwardCompatibilityService(EmailDeliveryInterface[EmailTemplateVars]):
    ep_backward_compatiblity_service: EPBackwardCompatibilityService

    def __init__(
        self,
        app_info: AppInfo,
        ep_recipe_interface_impl: EPRecipeInterface,
        reset_password_using_token_feature: Union[
            InputResetPasswordUsingTokenFeature, None
        ] = None,
    ) -> None:
        self.ep_backward_compatiblity_service = EPBackwardCompatibilityService(
            app_info,
            ep_recipe_interface_impl,
            reset_password_using_token_feature,
        )

    async def send_email(
        self, template_vars: EmailTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        await self.ep_backward_compatiblity_service.send_email(
            template_vars, user_context
        )



from ...interfaces import APIInterface, APIOptions, UserDeleteAPIResponse
from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python import Supertokens


async def handle_user_delete(
    _api_interface: APIInterface, api_options: APIOptions
) -> UserDeleteAPIResponse:
    user_id = api_options.request.get_query_param("userId")

    if user_id is None:
        raise_bad_input_exception("Missing required parameter 'userId'")

    await Supertokens.get_instance().delete_user(user_id)

    return UserDeleteAPIResponse()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, Dict, Union

from supertokens_python.ingredients.smsdelivery.types import (
    SMSContent,
    TwilioServiceInterface,
)
from supertokens_python.recipe.passwordless.smsdelivery.services.twilio.passwordless_login import (
    pless_sms_content,
)
from supertokens_python.recipe.passwordless.types import (
    PasswordlessLoginSMSTemplateVars,
)


class ServiceImplementation(TwilioServiceInterface[PasswordlessLoginSMSTemplateVars]):
    async def send_raw_sms(
        self,
        content: SMSContent,
        user_context: Dict[str, Any],
        from_: Union[str, None] = None,
        messaging_service_sid: Union[str, None] = None,
    ) -> None:
        if from_:
            self.twilio_client.messages.create(  # type: ignore
                to=content.to_phone,
                body=content.body,
                from_=from_,
            )
        else:
            self.twilio_client.messages.create(  # type: ignore
                to=content.to_phone,
                body=content.body,
                messaging_service_sid=messaging_service_sid,
            )

    async def get_content(
        self,
        template_vars: PasswordlessLoginSMSTemplateVars,
        user_context: Dict[str, Any],
    ) -> SMSContent:
        _ = user_context
        return pless_sms_content(template_vars)



from typing import Any, Dict, Optional, Union

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.dashboard.utils import (
    get_user_for_recipe_id,
    is_valid_recipe_id,
)
from supertokens_python.recipe.emailpassword import EmailPasswordRecipe
from supertokens_python.recipe.emailpassword.asyncio import (
    update_email_or_password as ep_update_email_or_password,
)
from supertokens_python.recipe.emailpassword.constants import FORM_FIELD_EMAIL_ID
from supertokens_python.recipe.emailpassword.interfaces import (
    UpdateEmailOrPasswordEmailAlreadyExistsError,
    UpdateEmailOrPasswordUnknownUserIdError,
)
from supertokens_python.recipe.passwordless import PasswordlessRecipe
from supertokens_python.recipe.passwordless.asyncio import (
    update_user as pless_update_user,
)
from supertokens_python.recipe.passwordless.interfaces import (
    UpdateUserEmailAlreadyExistsError as EmailAlreadyExistsErrorResponse,
)
from supertokens_python.recipe.passwordless.interfaces import (
    UpdateUserPhoneNumberAlreadyExistsError as PhoneNumberAlreadyExistsError,
)
from supertokens_python.recipe.passwordless.interfaces import (
    UpdateUserUnknownUserIdError as PlessUpdateUserUnknownUserIdError,
)
from supertokens_python.recipe.passwordless.utils import (
    ContactEmailOnlyConfig,
    ContactEmailOrPhoneConfig,
    ContactPhoneOnlyConfig,
    default_validate_email,
    default_validate_phone_number,
)
from supertokens_python.recipe.thirdpartyemailpassword import (
    ThirdPartyEmailPasswordRecipe,
)
from supertokens_python.recipe.thirdpartyemailpassword.asyncio import (
    update_email_or_password as tpep_update_email_or_password,
)
from supertokens_python.recipe.thirdpartypasswordless import (
    ThirdPartyPasswordlessRecipe,
)
from supertokens_python.recipe.usermetadata import UserMetadataRecipe
from supertokens_python.recipe.usermetadata.asyncio import update_user_metadata

from ...interfaces import (
    APIInterface,
    APIOptions,
    UserPutAPIEmailAlreadyExistsErrorResponse,
    UserPutAPIInvalidEmailErrorResponse,
    UserPutAPIInvalidPhoneErrorResponse,
    UserPutAPIOkResponse,
    UserPutPhoneAlreadyExistsAPIResponse,
)


async def update_email_for_recipe_id(
    recipe_id: str, user_id: str, email: str
) -> Union[
    UserPutAPIOkResponse,
    UserPutAPIInvalidEmailErrorResponse,
    UserPutAPIEmailAlreadyExistsErrorResponse,
]:
    validation_error: Optional[str] = None

    if recipe_id == "emailpassword":
        form_fields = (
            EmailPasswordRecipe.get_instance().config.sign_up_feature.form_fields
        )
        email_form_fields = [
            form_field
            for form_field in form_fields
            if form_field.id == FORM_FIELD_EMAIL_ID
        ]

        validation_error = await email_form_fields[0].validate(email)

        if validation_error is not None:
            return UserPutAPIInvalidEmailErrorResponse(validation_error)

        email_update_response = await ep_update_email_or_password(user_id, email)

        if isinstance(
            email_update_response, UpdateEmailOrPasswordEmailAlreadyExistsError
        ):
            return UserPutAPIEmailAlreadyExistsErrorResponse()

        return UserPutAPIOkResponse()

    if recipe_id == "thirdpartyemailpassword":
        form_fields = (
            ThirdPartyEmailPasswordRecipe.get_instance().email_password_recipe.config.sign_up_feature.form_fields
        )
        email_form_fields = [
            form_field
            for form_field in form_fields
            if form_field.id == FORM_FIELD_EMAIL_ID
        ]

        validation_error = await email_form_fields[0].validate(email)

        if validation_error is not None:
            return UserPutAPIInvalidEmailErrorResponse(validation_error)

        email_update_response = await tpep_update_email_or_password(user_id, email)

        if isinstance(
            email_update_response, UpdateEmailOrPasswordEmailAlreadyExistsError
        ):
            return UserPutAPIEmailAlreadyExistsErrorResponse()

        if isinstance(email_update_response, UpdateEmailOrPasswordUnknownUserIdError):
            raise Exception("Should never come here")

        return UserPutAPIOkResponse()

    if recipe_id == "passwordless":
        validation_error = None

        passwordless_config = PasswordlessRecipe.get_instance().config.contact_config

        if isinstance(passwordless_config.contact_method, ContactPhoneOnlyConfig):
            validation_error = await default_validate_email(email)

        elif isinstance(
            passwordless_config, (ContactEmailOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            validation_error = await passwordless_config.validate_email_address(email)

        if validation_error is not None:
            return UserPutAPIInvalidEmailErrorResponse(validation_error)

        update_result = await pless_update_user(user_id, email)

        if isinstance(update_result, PlessUpdateUserUnknownUserIdError):
            raise Exception("Should never come here")

        if isinstance(update_result, EmailAlreadyExistsErrorResponse):
            return UserPutAPIEmailAlreadyExistsErrorResponse()

        return UserPutAPIOkResponse()

    if recipe_id == "thirdpartypasswordless":
        validation_error = None

        passwordless_config = (
            ThirdPartyPasswordlessRecipe.get_instance().passwordless_recipe.config.contact_config
        )

        if isinstance(passwordless_config, ContactPhoneOnlyConfig):
            validation_error = await default_validate_email(email)
        elif isinstance(
            passwordless_config, (ContactEmailOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            validation_error = await passwordless_config.validate_email_address(email)

        if validation_error is not None:
            return UserPutAPIInvalidEmailErrorResponse(validation_error)

        update_result = await pless_update_user(user_id, email)

        if isinstance(update_result, PlessUpdateUserUnknownUserIdError):
            raise Exception("Should never come here")

        if isinstance(update_result, EmailAlreadyExistsErrorResponse):
            return UserPutAPIEmailAlreadyExistsErrorResponse()

        return UserPutAPIOkResponse()

    # If it comes here then the user is a third party user in which case the UI should not have allowed this
    raise Exception("Should never come here")


async def update_phone_for_recipe_id(
    recipe_id: str, user_id: str, phone: str
) -> Union[
    UserPutAPIOkResponse,
    UserPutAPIInvalidPhoneErrorResponse,
    UserPutPhoneAlreadyExistsAPIResponse,
]:
    validation_error: Optional[str] = None

    if recipe_id == "passwordless":
        validation_error = None

        passwordless_config = PasswordlessRecipe.get_instance().config.contact_config

        if isinstance(passwordless_config, ContactEmailOnlyConfig):
            validation_error = await default_validate_phone_number(phone)
        elif isinstance(
            passwordless_config, (ContactPhoneOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            validation_error = await passwordless_config.validate_phone_number(phone)

        if validation_error is not None:
            return UserPutAPIInvalidPhoneErrorResponse(validation_error)

        update_result = await pless_update_user(user_id, phone_number=phone)

        if isinstance(update_result, PlessUpdateUserUnknownUserIdError):
            raise Exception("Should never come here")

        if isinstance(update_result, PhoneNumberAlreadyExistsError):
            return UserPutPhoneAlreadyExistsAPIResponse()

        return UserPutAPIOkResponse()

    if recipe_id == "thirdpartypasswordless":
        validation_error = None

        passwordless_config = (
            ThirdPartyPasswordlessRecipe.get_instance().passwordless_recipe.config.contact_config
        )

        if isinstance(passwordless_config, ContactEmailOnlyConfig):
            validation_error = await default_validate_phone_number(phone)

        elif isinstance(
            passwordless_config, (ContactPhoneOnlyConfig, ContactEmailOrPhoneConfig)
        ):
            validation_error = await passwordless_config.validate_phone_number(phone)

        if validation_error is not None:
            return UserPutAPIInvalidPhoneErrorResponse(validation_error)

        update_result = await pless_update_user(user_id, phone_number=phone)

        if isinstance(update_result, PlessUpdateUserUnknownUserIdError):
            raise Exception("Should never come here")

        if isinstance(update_result, PhoneNumberAlreadyExistsError):
            return UserPutPhoneAlreadyExistsAPIResponse()

        return UserPutAPIOkResponse()

    # If it comes here then the user is a third party user in which case the UI should not have allowed this
    raise Exception("Should never come here")


async def handle_user_put(
    _api_interface: APIInterface, api_options: APIOptions
) -> Union[
    UserPutAPIOkResponse,
    UserPutAPIInvalidEmailErrorResponse,
    UserPutAPIEmailAlreadyExistsErrorResponse,
    UserPutAPIInvalidPhoneErrorResponse,
    UserPutPhoneAlreadyExistsAPIResponse,
]:
    request_body: Dict[str, Any] = await api_options.request.json()  # type: ignore
    user_id: Optional[str] = request_body.get("userId")
    recipe_id: Optional[str] = request_body.get("recipeId")
    first_name: Optional[str] = request_body.get("firstName")
    last_name: Optional[str] = request_body.get("lastName")
    email: Optional[str] = request_body.get("email")
    phone: Optional[str] = request_body.get("phone")

    if not isinstance(user_id, str):
        return raise_bad_input_exception(
            "Required parameter 'userId' is missing or has an invalid type"
        )

    if not isinstance(recipe_id, str):
        return raise_bad_input_exception(
            "Required parameter 'recipeId' is missing or has an invalid type"
        )

    if not is_valid_recipe_id(recipe_id):
        raise_bad_input_exception("Invalid recipe id")

    if first_name is None and not isinstance(first_name, str):
        raise_bad_input_exception(
            "Required parameter 'firstName' is missing or has an invalid type"
        )

    if last_name is None and not isinstance(last_name, str):
        raise_bad_input_exception(
            "Required parameter 'lastName' is missing or has an invalid type"
        )

    if email is None and not isinstance(email, str):
        raise_bad_input_exception(
            "Required parameter 'email' is missing or has an invalid type"
        )

    if phone is None and not isinstance(phone, str):
        raise_bad_input_exception(
            "Required parameter 'phone' is missing or has an invalid type"
        )

    user_response = await get_user_for_recipe_id(user_id, recipe_id)

    if user_response is None:
        raise Exception("Should never come here")

    first_name = first_name.strip()
    last_name = last_name.strip()
    email = email.strip()
    phone = phone.strip()

    if first_name != "" or last_name != "":
        is_recipe_initialized = False

        try:
            UserMetadataRecipe.get_instance()
            is_recipe_initialized = True
        except Exception:
            pass

        if is_recipe_initialized:
            metadata_update = {}

            if first_name != "":
                metadata_update["first_name"] = first_name

            if last_name != "":
                metadata_update["last_name"] = last_name

            await update_user_metadata(user_id, metadata_update)

    if email != "":
        email_update_response = await update_email_for_recipe_id(
            user_response.recipe, user_id, email
        )

        if not isinstance(email_update_response, UserPutAPIOkResponse):
            return email_update_response

    if phone != "":
        phone_update_response = await update_phone_for_recipe_id(
            user_response.recipe, user_id, phone
        )

        if not isinstance(phone_update_response, UserPutAPIOkResponse):
            return phone_update_response

    return UserPutAPIOkResponse()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional, Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.types import APIResponse, GeneralErrorResponse

from ...supertokens import AppInfo
from ..session.interfaces import SessionContainer

if TYPE_CHECKING:
    from supertokens_python.framework import BaseRequest, BaseResponse

    from .types import User, VerificationEmailTemplateVars
    from .utils import EmailVerificationConfig


class CreateEmailVerificationTokenOkResult:
    def __init__(self, token: str):
        self.token = token


class CreateEmailVerificationTokenEmailAlreadyVerifiedError:
    pass


class VerifyEmailUsingTokenOkResult:
    def __init__(self, user: User):
        self.user = user


class VerifyEmailUsingTokenInvalidTokenError:
    pass


class RevokeEmailVerificationTokensOkResult:
    pass


class UnverifyEmailOkResult:
    pass


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def create_email_verification_token(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> Union[
        CreateEmailVerificationTokenOkResult,
        CreateEmailVerificationTokenEmailAlreadyVerifiedError,
    ]:
        pass

    @abstractmethod
    async def verify_email_using_token(
        self, token: str, user_context: Dict[str, Any]
    ) -> Union[VerifyEmailUsingTokenOkResult, VerifyEmailUsingTokenInvalidTokenError]:
        pass

    @abstractmethod
    async def is_email_verified(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> bool:
        pass

    @abstractmethod
    async def revoke_email_verification_tokens(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> RevokeEmailVerificationTokensOkResult:
        pass

    @abstractmethod
    async def unverify_email(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> UnverifyEmailOkResult:
        pass


class APIOptions:
    def __init__(
        self,
        request: BaseRequest,
        response: BaseResponse,
        recipe_id: str,
        config: EmailVerificationConfig,
        recipe_implementation: RecipeInterface,
        app_info: AppInfo,
        email_delivery: EmailDeliveryIngredient[VerificationEmailTemplateVars],
    ):
        self.request = request
        self.response = response
        self.recipe_id = recipe_id
        self.config = config
        self.recipe_implementation = recipe_implementation
        self.app_info = app_info
        self.email_delivery = email_delivery


class EmailVerifyPostOkResult(APIResponse):
    def __init__(self, user: User):
        self.user = user
        self.status = "OK"

    def to_json(self) -> Dict[str, Any]:
        return {
            "status": self.status,
            "user": {"id": self.user.user_id, "email": self.user.email},
        }


class EmailVerifyPostInvalidTokenError(APIResponse):
    def __init__(self):
        self.status = "EMAIL_VERIFICATION_INVALID_TOKEN_ERROR"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class IsEmailVerifiedGetOkResult(APIResponse):
    def __init__(self, is_verified: bool):
        self.status = "OK"
        self.is_verified = is_verified

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status, "isVerified": self.is_verified}


class GenerateEmailVerifyTokenPostOkResult(APIResponse):
    def __init__(self):
        self.status = "OK"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class GenerateEmailVerifyTokenPostEmailAlreadyVerifiedError(APIResponse):
    def __init__(self):
        self.status = "EMAIL_ALREADY_VERIFIED_ERROR"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class APIInterface(ABC):
    def __init__(self):
        self.disable_email_verify_post = False
        self.disable_is_email_verified_get = False
        self.disable_generate_email_verify_token_post = False

    @abstractmethod
    async def email_verify_post(
        self,
        token: str,
        session: Optional[SessionContainer],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        EmailVerifyPostOkResult, EmailVerifyPostInvalidTokenError, GeneralErrorResponse
    ]:
        pass

    @abstractmethod
    async def is_email_verified_get(
        self,
        session: SessionContainer,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[IsEmailVerifiedGetOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def generate_email_verify_token_post(
        self,
        session: SessionContainer,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        GenerateEmailVerifyTokenPostOkResult,
        GenerateEmailVerifyTokenPostEmailAlreadyVerifiedError,
        GeneralErrorResponse,
    ]:
        pass


class GetEmailForUserIdOkResult:
    def __init__(self, email: str):
        self.email = email


class EmailDoesNotExistError(Exception):
    pass


class UnknownUserIdError(Exception):
    pass


TypeGetEmailForUserIdFunction = Callable[
    [str, Dict[str, Any]],
    Awaitable[
        Union[GetEmailForUserIdOkResult, EmailDoesNotExistError, UnknownUserIdError]
    ],
]



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier
from supertokens_python.recipe_module import APIHandled, RecipeModule

from .api.implementation import APIImplementation
from .interfaces import APIInterface, APIOptions, RecipeInterface
from .recipe_implementation import RecipeImplementation
from ..emailverification.interfaces import GetEmailForUserIdOkResult, UnknownUserIdError
from ...post_init_callbacks import PostSTInitCallbacks

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from supertokens_python.framework.response import BaseResponse
    from supertokens_python.supertokens import AppInfo
    from .utils import SignInAndUpFeature, InputOverrideConfig

from supertokens_python.exceptions import SuperTokensError, raise_general_exception
from supertokens_python.recipe.emailverification.recipe import EmailVerificationRecipe

from .api import (
    handle_apple_redirect_api,
    handle_authorisation_url_api,
    handle_sign_in_up_api,
)
from .constants import APPLE_REDIRECT_HANDLER, AUTHORISATIONURL, SIGNINUP
from .exceptions import SuperTokensThirdPartyError
from .types import ThirdPartyIngredients
from .utils import validate_and_normalise_user_input


class ThirdPartyRecipe(RecipeModule):
    recipe_id = "thirdparty"
    __instance = None

    def __init__(
        self,
        recipe_id: str,
        app_info: AppInfo,
        sign_in_and_up_feature: SignInAndUpFeature,
        _ingredients: ThirdPartyIngredients,
        override: Union[InputOverrideConfig, None] = None,
    ):
        super().__init__(recipe_id, app_info)
        self.config = validate_and_normalise_user_input(
            sign_in_and_up_feature,
            override,
        )
        self.providers = self.config.sign_in_and_up_feature.providers
        recipe_implementation = RecipeImplementation(Querier.get_instance(recipe_id))
        self.recipe_implementation: RecipeInterface = (
            recipe_implementation
            if self.config.override.functions is None
            else self.config.override.functions(recipe_implementation)
        )
        api_implementation = APIImplementation()
        self.api_implementation: APIInterface = (
            api_implementation
            if self.config.override.apis is None
            else self.config.override.apis(api_implementation)
        )

        def callback():
            ev_recipe = EmailVerificationRecipe.get_instance_optional()
            if ev_recipe:
                ev_recipe.add_get_email_for_user_id_func(self.get_email_for_user_id)

        PostSTInitCallbacks.add_post_init_callback(callback)

    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        return isinstance(err, SuperTokensError) and (
            isinstance(err, SuperTokensThirdPartyError)
        )

    def get_apis_handled(self) -> List[APIHandled]:
        return [
            APIHandled(
                NormalisedURLPath(SIGNINUP),
                "post",
                SIGNINUP,
                self.api_implementation.disable_sign_in_up_post,
            ),
            APIHandled(
                NormalisedURLPath(AUTHORISATIONURL),
                "get",
                AUTHORISATIONURL,
                self.api_implementation.disable_authorisation_url_get,
            ),
            APIHandled(
                NormalisedURLPath(APPLE_REDIRECT_HANDLER),
                "post",
                APPLE_REDIRECT_HANDLER,
                self.api_implementation.disable_apple_redirect_handler_post,
            ),
        ]

    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ):
        api_options = APIOptions(
            request,
            response,
            self.recipe_id,
            self.config,
            self.recipe_implementation,
            self.providers,
            self.app_info,
        )

        if request_id == SIGNINUP:
            return await handle_sign_in_up_api(self.api_implementation, api_options)
        if request_id == AUTHORISATIONURL:
            return await handle_authorisation_url_api(
                self.api_implementation, api_options
            )
        if request_id == APPLE_REDIRECT_HANDLER:
            return await handle_apple_redirect_api(self.api_implementation, api_options)

        return None

    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:  # type: ignore
        if isinstance(err, SuperTokensThirdPartyError):
            raise err

    def get_all_cors_headers(self) -> List[str]:
        return []

    @staticmethod
    def init(
        sign_in_and_up_feature: SignInAndUpFeature,
        override: Union[InputOverrideConfig, None] = None,
    ):
        def func(app_info: AppInfo):
            if ThirdPartyRecipe.__instance is None:
                ingredients = ThirdPartyIngredients()
                ThirdPartyRecipe.__instance = ThirdPartyRecipe(
                    ThirdPartyRecipe.recipe_id,
                    app_info,
                    sign_in_and_up_feature,
                    ingredients,
                    override,
                )
                return ThirdPartyRecipe.__instance
            raise_general_exception(
                "ThirdParty recipe has already been initialised. Please check your code for bugs."
            )

        return func

    @staticmethod
    def get_instance() -> ThirdPartyRecipe:
        if ThirdPartyRecipe.__instance is not None:
            return ThirdPartyRecipe.__instance
        raise_general_exception(
            "Initialisation not done. Did you forget to call the SuperTokens.init function?"
        )

    @staticmethod
    def reset():
        if ("SUPERTOKENS_ENV" not in environ) or (
            environ["SUPERTOKENS_ENV"] != "testing"
        ):
            raise_general_exception("calling testing function in non testing env")
        ThirdPartyRecipe.__instance = None

    # instance functions below...............

    async def get_email_for_user_id(self, user_id: str, user_context: Dict[str, Any]):
        user_info = await self.recipe_implementation.get_user_by_id(
            user_id, user_context
        )
        if user_info is not None:
            return GetEmailForUserIdOkResult(user_info.email)

        return UnknownUserIdError()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from typing import Union

from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.asyncio import delete_user, get_user_count
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.querier import Querier
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.interfaces import APIInterface
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from supertokens_python.utils import is_version_gte
from tests.utils import (
    clean_st,
    extract_all_cookies,
    reset,
    setup_st,
    sign_up_request,
    start_st,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/refresh")
    async def custom_refresh(request: Request):  # type: ignore
        await refresh_session(request)
        return {}  # type: ignore

    @app.get("/info")
    async def info_get(request: Request):  # type: ignore
        await get_session(request, True)
        return {}  # type: ignore

    @app.get("/custom/info")
    def custom_info(_):  # type: ignore
        return {}  # type: ignore

    @app.options("/custom/handle")
    def custom_handle_options(_):  # type: ignore
        return {"method": "option"}

    @app.get("/handle")
    async def handle_get(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return {"s": session.get_handle()}

    @app.post("/logout")
    async def custom_logout(request: Request):  # type: ignore
        session: Union[None, SessionContainer] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        await session.revoke_session()
        return {}  # type: ignore

    return TestClient(app)


@mark.asyncio
async def test_that_disabling_api_the_default_signin_API_does_not_work(
    driver_config_client: TestClient,
):
    def apis_override_email_password(param: APIInterface):
        param.disable_sign_in_post = True
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=apis_override_email_password
                )
            )
        ],
    )
    start_st()

    response_1 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "random@gmail.com"},
            ]
        },
    )

    assert response_1.status_code == 404


@mark.asyncio
async def test_singinAPI_works_when_input_is_fine(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    user_info = dict_response["user"]
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "random@gmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["user"]["id"] == user_info["id"]
    assert dict_response["user"]["email"] == user_info["email"]


@mark.asyncio
async def test_singinAPI_throws_an_error_when_email_does_not_match(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "ra@gmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "WRONG_CREDENTIALS_ERROR"


@mark.asyncio
async def test_singin_api_throws_an_error_when_email_does_not_match(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "ra@gmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "WRONG_CREDENTIALS_ERROR"


@mark.asyncio
async def test_singinAPI_throws_an_error_if_password_is_incorrect(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "wrong_password"},
                {"id": "email", "value": "random@gmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "WRONG_CREDENTIALS_ERROR"


@mark.asyncio
async def test_bad_input_not_a_JSON_to_signin_api(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(url="/auth/signin", json={"x": "y"})

    assert response_2.status_code == 400
    dict_response = json.loads(response_2.text)
    assert dict_response["message"] == "Are you sending too many / too few formFields?"


@mark.asyncio
async def test_bad_input_not_a_JSON_to_signin_API(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(url="/auth/signin", json={"x": "y"})

    assert response_2.status_code == 400
    dict_response = json.loads(response_2.text)
    assert dict_response["message"] == "Are you sending too many / too few formFields?"


@mark.asyncio
async def test_that_a_successful_signin_yields_a_session(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpass123"},
                {"id": "email", "value": "random@gmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200

    cookies = extract_all_cookies(response_1)

    assert cookies["sAccessToken"] is not None
    assert cookies["sRefreshToken"] is not None
    assert response_2.headers.get("anti-csrf") is not None


@mark.asyncio
async def test_email_field_validation_error(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpassword123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={
            "formFields": [
                {"id": "password", "value": "validpassword123"},
                {"id": "email", "value": "randomgmail.com"},
            ]
        },
    )

    assert response_2.status_code == 200
    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "FIELD_ERROR"


@mark.asyncio
async def test_formFields_has_no_email_field(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpassword123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={"formFields": [{"id": "password", "value": "validpassword123"}]},
    )

    assert response_2.status_code == 400


@mark.asyncio
async def test_formFields_has_no_password_field(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpassword123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    response_2 = driver_config_client.post(
        url="/auth/signin",
        json={"formFields": [{"id": "email", "value": "randomgmail.com"}]},
    )

    assert response_2.status_code == 400


@mark.asyncio
async def test_delete_user(driver_config_client: TestClient):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[emailpassword.init(), session.init(anti_csrf="VIA_TOKEN")],
    )
    start_st()

    response_1 = sign_up_request(
        driver_config_client, "random@gmail.com", "validpass123"
    )
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    user_count = await get_user_count()
    assert user_count == 1

    version = await Querier.get_instance().get_api_version()

    if is_version_gte(version, "2.10"):
        await delete_user(dict_response["user"]["id"])
        user_count = await get_user_count()
        assert user_count == 0
    else:
        error_raised = True
        try:
            await delete_user(dict_response["user"]["id"])
            error_raised = False
        except Exception:
            pass

        assert error_raised


# TODO add few more tests



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, Optional, Union

from supertokens_python.async_to_sync_wrapper import sync
from supertokens_python.recipe.emailverification.types import EmailTemplateVars


def create_email_verification_token(
    user_id: str,
    email: Optional[str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from supertokens_python.recipe.emailverification.asyncio import (
        create_email_verification_token,
    )

    return sync(create_email_verification_token(user_id, email, user_context))


def verify_email_using_token(
    token: str, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailverification.asyncio import (
        verify_email_using_token,
    )

    return sync(verify_email_using_token(token, user_context))


def is_email_verified(
    user_id: str,
    email: Optional[str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from supertokens_python.recipe.emailverification.asyncio import is_email_verified

    return sync(is_email_verified(user_id, email, user_context))


def revoke_email_verification_tokens(
    user_id: str,
    email: Optional[str] = None,
    user_context: Optional[Dict[str, Any]] = None,
):
    from supertokens_python.recipe.emailverification.asyncio import (
        revoke_email_verification_tokens,
    )

    return sync(revoke_email_verification_tokens(user_id, email, user_context))


def unverify_email(
    user_id: str,
    email: Optional[str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from supertokens_python.recipe.emailverification.asyncio import unverify_email

    return sync(unverify_email(user_id, email, user_context))


def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailverification.asyncio import send_email

    return sync(send_email(input_, user_context))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supertokens_python.recipe.thirdparty.interfaces import APIInterface, APIOptions

from supertokens_python.utils import default_user_context


async def handle_apple_redirect_api(
    api_implementation: APIInterface, api_options: APIOptions
):
    if api_implementation.disable_apple_redirect_handler_post:
        return None
    body = await api_options.request.form_data()

    code = body["code"] if "code" in body else ""
    state = body["state"] if "state" in body else ""
    user_context = default_user_context(api_options.request)

    # this will redirect the user...
    await api_implementation.apple_redirect_handler_post(
        code, state, api_options, user_context
    )

    return api_options.response



from unittest.mock import MagicMock, patch

from pytest import mark
from supertokens_python import init
from supertokens_python.framework import BaseRequest
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session_information,
    remove_claim,
)
from supertokens_python.recipe.session.session_class import Session
from tests.sessions.claims.utils import TrueClaim, get_st_init_args
from tests.utils import setup_function, start_st, teardown_function

from tests.utils import AsyncMock

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore

pytestmark = mark.asyncio


async def test_should_attempt_to_set_claim_to_none():
    recipe_implementation_mock = AsyncMock()
    session_config_mock = MagicMock()
    session = Session(
        recipe_implementation_mock,
        session_config_mock,
        "test_access_token",
        "test_session_handle",
        "test_user_id",
        {},
        "cookie",
    )
    with patch.object(
        Session,
        "merge_into_access_token_payload",
        wraps=session.merge_into_access_token_payload,
    ) as mock:
        await session.remove_claim(TrueClaim)
        mock.assert_called_once_with({"st-true": None}, {})


async def test_should_clear_previously_set_claim(timestamp: int):
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s: SessionContainer = await create_new_session(dummy_req, "someId")

    payload = s.get_access_token_payload()

    assert payload == {"st-true": {"v": True, "t": timestamp}}


async def test_should_clear_previously_set_claim_using_handle(timestamp: int):
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    dummy_req: BaseRequest = MagicMock()
    s: SessionContainer = await create_new_session(dummy_req, "someId")

    payload = s.get_access_token_payload()
    assert payload == {"st-true": {"v": True, "t": timestamp}}

    res = await remove_claim(s.get_handle(), TrueClaim)
    assert res is True

    session_info = await get_session_information(s.get_handle())
    assert session_info is not None
    payload_after = session_info.access_token_payload
    assert payload_after == {}


async def test_should_work_ok_for_non_existing_handle():
    init(**get_st_init_args(TrueClaim))  # type:ignore
    start_st()

    res = await remove_claim("non-existing-handle", TrueClaim)
    assert res is False



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supertokens_python.recipe.dashboard.interfaces import APIInterface, APIOptions

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier

from ..interfaces import SignOutOK


async def handle_emailpassword_signout_api(
    _: APIInterface, api_options: APIOptions
) -> SignOutOK:
    if api_options.config.auth_mode == "api-key":
        return SignOutOK()
    session_id_form_auth_header = api_options.request.get_header("authorization")
    if not session_id_form_auth_header:
        return raise_bad_input_exception(
            "Neither 'API Key' nor 'Authorization' header was found"
        )
    session_id_form_auth_header = session_id_form_auth_header.split()[1]
    await Querier.get_instance().send_delete_request(
        NormalisedURLPath("/recipe/dashboard/session"),
        {"sessionId": session_id_form_auth_header},
    )
    return SignOutOK()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations


from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.emailverification.interfaces import (
    APIInterface,
    APIOptions,
)
from supertokens_python.utils import (
    default_user_context,
    normalise_http_method,
    send_200_response,
)
from supertokens_python.recipe.session.asyncio import get_session


async def handle_email_verify_api(
    api_implementation: APIInterface, api_options: APIOptions
):
    user_context = default_user_context(api_options.request)
    if normalise_http_method(api_options.request.method()) == "post":
        if api_implementation.disable_email_verify_post:
            return None
        body = await api_options.request.json()
        if body is None:
            raise_bad_input_exception("Please pass JSON input body")
        if "token" not in body:
            raise_bad_input_exception("Please provide the email verification token")
        if not isinstance(body["token"], str):
            raise_bad_input_exception("The email verification token must be a string")

        token = body["token"]

        session = await get_session(
            api_options.request,
            session_required=False,
            override_global_claim_validators=lambda _, __, ___: [],
            user_context=user_context,
        )

        result = await api_implementation.email_verify_post(
            token, session, api_options, user_context
        )
    else:
        if api_implementation.disable_is_email_verified_get:
            return None

        session = await get_session(
            api_options.request,
            override_global_claim_validators=lambda _, __, ___: [],
            user_context=user_context,
        )
        assert session is not None
        result = await api_implementation.is_email_verified_get(
            session, api_options, user_context
        )

    return send_200_response(result.to_json(), api_options.response)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import asyncio
import base64
import json
from typing import Any, Dict, Optional, Union

from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark, skip
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.asyncio import delete_user
from supertokens_python.exceptions import BadInputError
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.querier import Querier
from supertokens_python.recipe import emailpassword, emailverification, session
from supertokens_python.recipe.emailverification.asyncio import (
    create_email_verification_token,
    is_email_verified,
    revoke_email_verification_tokens,
    unverify_email,
    verify_email_using_token,
)
from supertokens_python.recipe.emailverification.interfaces import (
    APIInterface,
    APIOptions,
    CreateEmailVerificationTokenOkResult,
    EmailVerifyPostOkResult,
    VerifyEmailUsingTokenInvalidTokenError,
)
from supertokens_python.recipe.emailverification.types import User as EVUser
from supertokens_python.recipe.emailverification.utils import OverrideConfig
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import (
    create_new_session,
    get_session,
    refresh_session,
)
from supertokens_python.recipe.session.constants import ANTI_CSRF_HEADER_KEY
from supertokens_python.utils import is_version_gte
from tests.utils import (
    TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY,
    email_verify_token_request,
    extract_all_cookies,
    get_st_init_args,
    min_api_version,
    set_key_value_in_config,
    setup_function,
    sign_up_request,
    start_st,
    teardown_function,
)

_ = setup_function  # type: ignore
_ = teardown_function  # type: ignore

pytestmark = mark.asyncio


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/login")
    async def login(request: Request):  # type: ignore
        user_id = "userId"
        await create_new_session(request, user_id, {}, {})
        return {"userId": user_id}

    @app.post("/refresh")
    async def custom_refresh(request: Request):  # type: ignore
        await refresh_session(request)
        return {}  # type: ignore

    @app.get("/info")
    async def info_get(request: Request):  # type: ignore
        await get_session(request, True)
        return {}  # type: ignore

    @app.get("/custom/info")
    def custom_info(_):  # type: ignore
        return {}  # type: ignore

    @app.options("/custom/handle")
    def custom_handle_options(_):  # type: ignore
        return {"method": "option"}

    @app.get("/handle")
    async def handle_get(request: Request):  # type: ignore
        session: Union[SessionContainer, None] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return {"s": session.get_handle()}

    @app.post("/logout")
    async def custom_logout(request: Request):  # type: ignore
        session: Union[SessionContainer, None] = await get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        await session.revoke_session()
        return {}  # type: ignore

    return TestClient(app)


async def test_the_generate_token_api_with_valid_input_email_not_verified(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    user_id = dict_response["user"]["id"]
    cookies = extract_all_cookies(response_1)

    response = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    dict_response = json.loads(response.text)
    assert dict_response["status"] == "OK"


async def test_the_generate_token_api_with_valid_input_email_verified_and_test_error(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")

    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"

    user_id = dict_response["user"]["id"]
    cookies = extract_all_cookies(response_1)

    verify_token = await create_email_verification_token(user_id)
    if isinstance(verify_token, CreateEmailVerificationTokenOkResult):
        await verify_email_using_token(verify_token.token)

        response = email_verify_token_request(
            driver_config_client,
            cookies["sAccessToken"]["value"],
            response_1.headers.get("anti-csrf"),  # type: ignore
            user_id,
        )
        dict_response = json.loads(response.text)
        assert dict_response["status"] == "EMAIL_ALREADY_VERIFIED_ERROR"
        return
    raise Exception("Test failed")


async def test_the_generate_token_api_with_valid_input_no_session_and_check_output(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = driver_config_client.post(url="/auth/user/email/verify/token")
    assert response_1.status_code == 401
    dict_response = json.loads(response_1.text)
    assert dict_response["message"] == "unauthorised"


async def test_the_generate_token_api_with_an_expired_access_token_and_see_that_try_refresh_token_is_returned(
    driver_config_client: TestClient,
):
    set_key_value_in_config(TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY, "2")

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    await asyncio.sleep(5)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    dict_response = json.loads(response_2.text)

    assert response_2.status_code == 401
    assert dict_response["message"] == "try refresh token"

    response_3 = driver_config_client.post(
        url="/auth/session/refresh",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
    )

    assert response_3.status_code == 200

    cookies1 = extract_all_cookies(response_3)
    response_4 = email_verify_token_request(
        driver_config_client,
        cookies1["sAccessToken"]["value"],
        response_3.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )

    dict_response = json.loads(response_4.text)

    assert response_4.status_code == 200
    assert dict_response["status"] == "OK"


async def test_that_providing_your_own_email_callback_and_make_sure_it_is_called(
    driver_config_client: TestClient,
):
    user_info: Union[None, EVUser] = None
    email_token = None

    async def custom_f(
        user: EVUser, email_verification_url_token: str, _: Dict[str, Any]
    ):
        nonlocal user_info, email_token
        user_info = user
        email_token = email_verification_url_token

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED", create_and_send_custom_email=custom_f
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id: str = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"
    if user_info is None:
        raise Exception("Should never come here")
    assert user_info.user_id == user_id  # type: ignore
    assert user_info.email == "test@gmail.com"  # type: ignore
    assert email_token is not None


async def test_the_email_verify_api_with_valid_input(driver_config_client: TestClient):
    token = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED", create_and_send_custom_email=custom_f
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        json={"method": "token", "token": token},
    )

    dict_response = json.loads(response_3.text)
    assert dict_response["status"] == "OK"


async def test_the_email_verify_api_with_invalid_token_and_check_error(
    driver_config_client: TestClient,
):
    token = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED", create_and_send_custom_email=custom_f
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        json={"method": "token", "token": "bad token"},
    )

    dict_response = json.loads(response_3.text)
    assert dict_response["status"] == "EMAIL_VERIFICATION_INVALID_TOKEN_ERROR"


async def test_the_email_verify_api_with_token_of_not_type_string(
    driver_config_client: TestClient,
):
    token = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED", create_and_send_custom_email=custom_f
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        json={"method": "token", "token": 200},
    )

    dict_response = json.loads(response_3.text)
    assert response_3.status_code == 400
    assert dict_response["message"] == "The email verification token must be a string"


async def test_that_the_handle_post_email_verification_callback_is_called_on_successful_verification_if_given(
    driver_config_client: TestClient,
):
    token = None
    user_info_from_callback: Union[None, EVUser] = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    def apis_override_email_password(param: APIInterface):
        temp = param.email_verify_post

        async def email_verify_post(
            token: str,
            session: Optional[SessionContainer],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ):
            nonlocal user_info_from_callback

            response = await temp(token, session, api_options, user_context)

            if isinstance(response, EmailVerifyPostOkResult):
                user_info_from_callback = response.user

            return response

        param.email_verify_post = email_verify_post
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED",
                create_and_send_custom_email=custom_f,
                override=OverrideConfig(apis=apis_override_email_password),
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        json={"method": "token", "token": token},
    )

    dict_response = json.loads(response_3.text)
    assert response_3.status_code == 200
    assert dict_response["status"] == "OK"

    await asyncio.sleep(1)
    if user_info_from_callback is None:
        raise Exception("Should never come here")
    assert user_info_from_callback.user_id == user_id  # type: ignore
    assert user_info_from_callback.email == "test@gmail.com"  # type: ignore


async def test_the_email_verify_with_valid_input_using_the_get_method(
    driver_config_client: TestClient,
):
    token = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED",
                create_and_send_custom_email=custom_f,
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        json={"method": "token", "token": token},
        headers={ANTI_CSRF_HEADER_KEY: response_1.headers.get("anti-csrf")},
    )

    dict_response = json.loads(response_3.text)
    assert response_3.status_code == 200
    assert dict_response["status"] == "OK"

    response_4 = driver_config_client.get(
        url="/auth/user/email/verify",
        headers={
            "Content-Type": "application/json",
            "anti-csrf": response_1.headers.get("anti-csrf"),
        },
        cookies={
            "sAccessToken": cookies["sAccessToken"]["value"],
        },
    )

    dict_response = json.loads(response_4.text)
    assert response_4.status_code == 200
    assert dict_response["status"] == "OK"


async def test_the_email_verify_with_no_session_using_the_get_method(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    response_4 = driver_config_client.get(url="/auth/user/email/verify")

    dict_response = json.loads(response_4.text)
    assert response_4.status_code == 401
    assert dict_response["message"] == "unauthorised"


async def test_the_email_verify_api_with_valid_input_overriding_apis(
    driver_config_client: TestClient,
):
    token = None
    user_info_from_callback: Union[None, EVUser] = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    def apis_override_email_password(param: APIInterface):
        temp = param.email_verify_post

        async def email_verify_post(
            token: str,
            session: Optional[SessionContainer],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ):
            nonlocal user_info_from_callback

            response = await temp(token, session, api_options, user_context)

            if isinstance(response, EmailVerifyPostOkResult):
                user_info_from_callback = response.user

            return response

        param.email_verify_post = email_verify_post
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED",
                create_and_send_custom_email=custom_f,
                override=OverrideConfig(apis=apis_override_email_password),
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        json={"method": "token", "token": token},
        headers={ANTI_CSRF_HEADER_KEY: response_1.headers.get("anti-csrf")},
    )

    dict_response = json.loads(response_3.text)
    assert response_3.status_code == 200
    assert dict_response["status"] == "OK"

    await asyncio.sleep(1)

    if user_info_from_callback is None:
        raise Exception("Should never come here")
    assert user_info_from_callback.user_id == user_id  # type: ignore
    assert user_info_from_callback.email == "test@gmail.com"  # type: ignore


async def test_the_email_verify_api_with_valid_input_overriding_apis_throws_error(
    driver_config_client: TestClient,
):
    token = None
    user_info_from_callback: Union[None, EVUser] = None

    async def custom_f(
        _: EVUser, email_verification_url_token: str, __: Dict[str, Any]
    ):
        nonlocal token
        token = email_verification_url_token.split("?token=")[1].split("&rid=")[0]

    def apis_override_email_password(param: APIInterface):
        temp = param.email_verify_post

        async def email_verify_post(
            token: str,
            session: Optional[SessionContainer],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ):
            nonlocal user_info_from_callback

            response = await temp(token, session, api_options, user_context)

            if isinstance(response, EmailVerifyPostOkResult):
                user_info_from_callback = response.user

            raise BadInputError("verify exception")

        param.email_verify_post = email_verify_post
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(
                mode="REQUIRED",
                create_and_send_custom_email=custom_f,
                override=emailverification.InputOverrideConfig(
                    apis=apis_override_email_password
                ),
            ),
            emailpassword.init(),
        ],
    )
    start_st()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    await asyncio.sleep(1)
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    cookies = extract_all_cookies(response_1)

    response_2 = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        response_1.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    await asyncio.sleep(2)

    dict_response = json.loads(response_2.text)
    assert dict_response["status"] == "OK"

    assert token is not None

    response_3 = driver_config_client.post(
        url="/auth/user/email/verify",
        json={"method": "token", "token": token},
        headers={ANTI_CSRF_HEADER_KEY: response_1.headers.get("anti-csrf")},
    )

    dict_response = json.loads(response_3.text)
    assert response_3.status_code == 400
    assert dict_response["message"] == "verify exception"

    await asyncio.sleep(1)

    if user_info_from_callback is None:
        raise Exception("Should never come here")
    assert user_info_from_callback.user_id == user_id  # type: ignore
    assert user_info_from_callback.email == "test@gmail.com"  # type: ignore


async def test_the_generate_token_api_with_valid_input_and_then_remove_token(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init("OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.9"):
        # If the version less than 2.9, the recipe doesn't exist. So skip the test
        skip()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    verify_token = await create_email_verification_token(user_id)
    await revoke_email_verification_tokens(user_id)

    if isinstance(verify_token, CreateEmailVerificationTokenOkResult):
        response = await verify_email_using_token(verify_token.token)
        assert isinstance(response, VerifyEmailUsingTokenInvalidTokenError)
        return
    raise Exception("Test failed")


async def test_the_generate_token_api_with_valid_input_verify_and_then_unverify_email(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailverification.init(mode="OPTIONAL"),
            emailpassword.init(),
        ],
    )
    start_st()

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.9"):
        # If the version is less than 2.9, the recipe doesn't exist. So skip the test.
        skip()

    response_1 = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    assert response_1.status_code == 200
    dict_response = json.loads(response_1.text)
    assert dict_response["status"] == "OK"
    user_id = dict_response["user"]["id"]

    verify_token = await create_email_verification_token(user_id)
    if isinstance(verify_token, CreateEmailVerificationTokenOkResult):
        await verify_email_using_token(verify_token.token)

        assert await is_email_verified(user_id)

        await unverify_email(user_id)

        is_verified = await is_email_verified(user_id)
        assert is_verified is False
        return
    raise Exception("Test failed")


@min_api_version("2.11")
async def test_email_verify_with_deleted_user(driver_config_client: TestClient):
    async def custom_f(_: EVUser, __: str, ___: Optional[Dict[str, Any]]):
        return None

    st_args = get_st_init_args(
        [
            emailpassword.init(),
            emailverification.init("OPTIONAL", create_and_send_custom_email=custom_f),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ]
    )
    init(**st_args)
    start_st()

    res = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    dict_response = json.loads(res.text)

    assert res.status_code == 200
    assert dict_response["status"] == "OK"

    user_id = dict_response["user"]["id"]
    cookies = extract_all_cookies(res)

    await delete_user(user_id)

    response = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        res.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    dict_response = json.loads(response.text)

    assert response.status_code == 401
    assert dict_response == {"message": "unauthorised"}


@min_api_version("2.11")
async def test_generate_email_verification_token_api_updates_session_claims(
    driver_config_client: TestClient,
):
    async def custom_f(_: EVUser, __: str, ___: Optional[Dict[str, Any]]):
        return None

    init(
        **get_st_init_args(
            [
                emailpassword.init(),
                emailverification.init(
                    "OPTIONAL", create_and_send_custom_email=custom_f
                ),
                session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            ]
        )
    )
    start_st()

    # Create user:
    res = sign_up_request(driver_config_client, "test@gmail.com", "testPass123")
    dict_res = res.json()
    assert dict_res["status"] == "OK"
    assert res.status_code == 200

    user_id = dict_res["user"]["id"]
    cookies = extract_all_cookies(res)

    # Start verification:
    verify_token = await create_email_verification_token(user_id)
    assert isinstance(verify_token, CreateEmailVerificationTokenOkResult)
    await verify_email_using_token(verify_token.token)

    res = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        res.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    cookies = extract_all_cookies(res)
    assert res.status_code == 200
    assert res.json() == {"status": "EMAIL_ALREADY_VERIFIED_ERROR"}
    front_token = json.loads(
        base64.b64decode(res.headers["front-token"])
    )  # utf_base64decodes
    assert front_token["up"]["st-ev"]["v"] is True

    # calling the API again should not modify the access token again
    res = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        res.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    assert res.status_code == 200
    assert res.json() == {"status": "EMAIL_ALREADY_VERIFIED_ERROR"}
    assert "front-token" not in res.headers

    # now we mark the email as unverified and try again:
    await unverify_email(user_id)
    res = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        res.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    cookies = extract_all_cookies(res)
    assert res.status_code == 200
    assert res.json() == {"status": "OK"}
    front_token = json.loads(base64.b64decode(res.headers["front-token"]))
    assert front_token["up"]["st-ev"]["v"] is False

    # calling the API again should not modify the access token again:
    res = email_verify_token_request(
        driver_config_client,
        cookies["sAccessToken"]["value"],
        res.headers.get("anti-csrf"),  # type: ignore
        user_id,
    )
    assert res.status_code == 200
    assert res.json() == {"status": "OK"}
    assert "front-token" not in res.headers



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from fastapi import FastAPI
import json
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session
from tests.utils import (
    get_st_init_args,
    setup_function,
    start_st,
    teardown_function,
    sign_up_request,
)

_ = setup_function  # type: ignore
_ = teardown_function  # type: ignore

pytestmark = mark.asyncio


@fixture(scope="function")
async def app():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


async def test_field_error_on_existing_email_signup(app: TestClient):
    init_args = get_st_init_args([emailpassword.init(), session.init()])
    init(**init_args)
    start_st()

    response = json.loads(sign_up_request(app, "random@gmail.com", "validpass123").text)
    assert response["status"] == "OK"

    response = json.loads(sign_up_request(app, "random@gmail.com", "validpass123").text)
    assert response == {
        "status": "FIELD_ERROR",
        "formFields": [
            {
                "id": "email",
                "error": "This email already exists. Please sign in instead.",
            }
        ],
    }



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
from datetime import datetime
from os import getenv, path
from typing import Union

from .constants import VERSION

NAMESPACE = "com.supertokens"
DEBUG_ENV_VAR = "SUPERTOKENS_DEBUG"

supertokens_dir = path.dirname(__file__)

# Configure logger
_logger = logging.getLogger(NAMESPACE)
debug_env = getenv(DEBUG_ENV_VAR, "").lower()
if debug_env == "1":
    _logger.setLevel(logging.DEBUG)


def _get_log_timestamp() -> str:
    return datetime.utcnow().isoformat()[:-3] + "Z"


class CustomStreamHandler(logging.StreamHandler):  # type: ignore
    def emit(self, record: logging.LogRecord):
        relative_path = path.relpath(record.pathname, supertokens_dir)

        record.msg = json.dumps(
            {
                "t": _get_log_timestamp(),
                "sdkVer": VERSION,
                "message": record.msg,
                "file": f"{relative_path}:{record.lineno}",
            }
        )

        return super().emit(record)


# Add stream handler and format
streamHandler = CustomStreamHandler()
streamFormatter = logging.Formatter("{name} {message}\n", style="{")
streamHandler.setFormatter(streamFormatter)
_logger.addHandler(streamHandler)


# The debug logger can be used like this:
# log_debug_message("Hello")
# Output log format:
# com.supertokens {"t": "2022-03-24T06:28:33.659Z", "sdkVer": "0.5.1", "message": "Hello", "file": "logger.py:73"}

# Export logger.debug as log_debug_message function
log_debug_message = _logger.debug


def get_maybe_none_as_str(o: Union[str, None]) -> str:
    if o is None:
        return "None"
    return o



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict, List, Union

from supertokens_python.recipe.passwordless.interfaces import (
    ConsumeCodeExpiredUserInputCodeError,
    ConsumeCodeIncorrectUserInputCodeError,
    ConsumeCodeOkResult,
    ConsumeCodeRestartFlowError,
    CreateCodeOkResult,
    CreateNewCodeForDeviceOkResult,
    CreateNewCodeForDeviceRestartFlowError,
    CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    DeleteUserInfoOkResult,
    DeleteUserInfoUnknownUserIdError,
    DeviceType,
    RecipeInterface,
    RevokeAllCodesOkResult,
    RevokeCodeOkResult,
    UpdateUserEmailAlreadyExistsError,
    UpdateUserOkResult,
    UpdateUserPhoneNumberAlreadyExistsError,
    UpdateUserUnknownUserIdError,
)

from ...passwordless.types import User
from ..interfaces import ConsumeCodeOkResult as ThirdPartyConsumeCodeOkResult
from ..interfaces import RecipeInterface as ThirdPartyPasswordlessInterface


class RecipeImplementation(RecipeInterface):
    def __init__(self, recipe_implementation: ThirdPartyPasswordlessInterface):
        super().__init__()
        self.recipe_implementation = recipe_implementation

    async def create_code(
        self,
        email: Union[None, str],
        phone_number: Union[None, str],
        user_input_code: Union[None, str],
        user_context: Dict[str, Any],
    ) -> CreateCodeOkResult:
        return await self.recipe_implementation.create_code(
            email, phone_number, user_input_code, user_context
        )

    async def create_new_code_for_device(
        self,
        device_id: str,
        user_input_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        CreateNewCodeForDeviceOkResult,
        CreateNewCodeForDeviceRestartFlowError,
        CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    ]:
        return await self.recipe_implementation.create_new_code_for_device(
            device_id, user_input_code, user_context
        )

    async def consume_code(
        self,
        pre_auth_session_id: str,
        user_input_code: Union[str, None],
        device_id: Union[str, None],
        link_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        ConsumeCodeOkResult,
        ConsumeCodeIncorrectUserInputCodeError,
        ConsumeCodeExpiredUserInputCodeError,
        ConsumeCodeRestartFlowError,
    ]:
        result = await self.recipe_implementation.consume_code(
            pre_auth_session_id, user_input_code, device_id, link_code, user_context
        )
        if isinstance(result, ThirdPartyConsumeCodeOkResult):
            return ConsumeCodeOkResult(
                result.created_new_user,
                User(
                    result.user.user_id,
                    result.user.email,
                    result.user.phone_number,
                    result.user.time_joined,
                ),
            )
        return result

    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        otherTypeUser = await self.recipe_implementation.get_user_by_id(
            user_id, user_context
        )
        if otherTypeUser is not None:
            if otherTypeUser.third_party_info is None:
                return User(
                    otherTypeUser.user_id,
                    otherTypeUser.email,
                    otherTypeUser.phone_number,
                    otherTypeUser.time_joined,
                )

        return None

    async def get_user_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        users = await self.recipe_implementation.get_users_by_email(email, user_context)
        for user in users:
            if user.third_party_info is not None:
                return User(
                    user.user_id, user.email, user.phone_number, user.time_joined
                )

        return None

    async def get_user_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        otherTypeUser = await self.recipe_implementation.get_user_by_phone_number(
            phone_number, user_context
        )
        if otherTypeUser is not None:
            if otherTypeUser.third_party_info is None:
                return User(
                    otherTypeUser.user_id,
                    otherTypeUser.email,
                    otherTypeUser.phone_number,
                    otherTypeUser.time_joined,
                )

        return None

    async def update_user(
        self,
        user_id: str,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        UpdateUserOkResult,
        UpdateUserUnknownUserIdError,
        UpdateUserEmailAlreadyExistsError,
        UpdateUserPhoneNumberAlreadyExistsError,
    ]:
        return await self.recipe_implementation.update_passwordless_user(
            user_id, email, phone_number, user_context
        )

    async def delete_email_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
        return await self.recipe_implementation.delete_email_for_passwordless_user(
            user_id, user_context
        )

    async def delete_phone_number_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
        return await self.recipe_implementation.delete_phone_number_for_user(
            user_id, user_context
        )

    async def revoke_all_codes(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> RevokeAllCodesOkResult:
        return await self.recipe_implementation.revoke_all_codes(
            email, phone_number, user_context
        )

    async def revoke_code(
        self, code_id: str, user_context: Dict[str, Any]
    ) -> RevokeCodeOkResult:
        return await self.recipe_implementation.revoke_code(code_id, user_context)

    async def list_codes_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        return await self.recipe_implementation.list_codes_by_email(email, user_context)

    async def list_codes_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        return await self.recipe_implementation.list_codes_by_phone_number(
            phone_number, user_context
        )

    async def list_codes_by_device_id(
        self, device_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        return await self.recipe_implementation.list_codes_by_device_id(
            device_id, user_context
        )

    async def list_codes_by_pre_auth_session_id(
        self, pre_auth_session_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        return await self.recipe_implementation.list_codes_by_pre_auth_session_id(
            pre_auth_session_id, user_context
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from httpx import AsyncClient
from supertokens_python.recipe.thirdparty.provider import Provider
from supertokens_python.recipe.thirdparty.types import (
    AccessTokenAPI,
    AuthorisationRedirectAPI,
    UserInfo,
    UserInfoEmail,
)

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest


class Facebook(Provider):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        scope: Union[None, List[str]] = None,
        is_default: bool = False,
    ):
        super().__init__("facebook", is_default)
        default_scopes = ["email"]

        if scope is None:
            scope = default_scopes
        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = list(set(scope))
        self.access_token_api_url = "https://graph.facebook.com/v9.0/oauth/access_token"
        self.authorisation_redirect_url = "https://www.facebook.com/v9.0/dialog/oauth"

    async def get_profile_info(
        self, auth_code_response: Dict[str, Any], user_context: Dict[str, Any]
    ) -> UserInfo:
        access_token: str = auth_code_response["access_token"]

        params = {"access_token": access_token, "fields": "id,email", "format": "json"}
        async with AsyncClient() as client:
            response = await client.get(  # type:ignore
                url="https://graph.facebook.com/me", params=params
            )
            user_info = response.json()
            user_id = user_info["id"]
            if "email" not in user_info or user_info["email"] is None:
                return UserInfo(user_id)
            return UserInfo(user_id, UserInfoEmail(user_info["email"], True))

    def get_authorisation_redirect_api_info(
        self, user_context: Dict[str, Any]
    ) -> AuthorisationRedirectAPI:
        params: Dict[str, Union[Callable[[BaseRequest], str], str]] = {
            "scope": " ".join(self.scopes),
            "response_type": "code",
            "client_id": self.client_id,
        }
        return AuthorisationRedirectAPI(self.authorisation_redirect_url, params)

    def get_access_token_api_info(
        self,
        redirect_uri: str,
        auth_code_from_request: str,
        user_context: Dict[str, Any],
    ) -> AccessTokenAPI:
        params = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": auth_code_from_request,
            "redirect_uri": redirect_uri,
        }
        return AccessTokenAPI(self.access_token_api_url, params)

    def get_redirect_uri(self, user_context: Dict[str, Any]) -> Union[None, str]:
        return None

    def get_client_id(self, user_context: Dict[str, Any]) -> str:
        return self.client_id



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Dict, Union

from supertokens_python.logger import log_debug_message
from supertokens_python.utils import get_timestamp_ms

from .exceptions import raise_try_refresh_token_exception
from .jwt import ParsedJWTInfo, verify_jwt


def sanitize_string(s: Any) -> Union[str, None]:
    if s == "":
        return s

    if not isinstance(s, str):
        return None

    return s.strip()


def sanitize_number(n: Any) -> Union[Union[int, float], None]:
    _type = type(n)
    if _type == int or _type == float:  # pylint: disable=consider-using-in
        return n

    return None


def get_info_from_access_token(
    jwt_info: ParsedJWTInfo, jwt_signing_public_key: str, do_anti_csrf_check: bool
):
    try:
        verify_jwt(jwt_info, jwt_signing_public_key)
        payload = jwt_info.payload

        validate_access_token_structure(payload)

        session_handle = sanitize_string(payload.get("sessionHandle"))
        user_id = sanitize_string(payload.get("userId"))
        refresh_token_hash_1 = sanitize_string(payload.get("refreshTokenHash1"))
        parent_refresh_token_hash_1 = sanitize_string(
            payload.get("parentRefreshTokenHash1")
        )
        user_data = payload.get("userData")
        anti_csrf_token = sanitize_string(payload.get("antiCsrfToken"))
        expiry_time = sanitize_number(payload.get("expiryTime"))
        time_created = sanitize_number(payload.get("timeCreated"))

        if anti_csrf_token is None and do_anti_csrf_check:
            raise Exception("Access token does not contain the anti-csrf token")

        assert isinstance(expiry_time, int)

        if expiry_time < get_timestamp_ms():
            raise Exception("Access token expired")

        return {
            "sessionHandle": session_handle,
            "userId": user_id,
            "refreshTokenHash1": refresh_token_hash_1,
            "parentRefreshTokenHash1": parent_refresh_token_hash_1,
            "userData": user_data,
            "antiCsrfToken": anti_csrf_token,
            "expiryTime": expiry_time,
            "timeCreated": time_created,
        }
    except Exception as e:
        log_debug_message(
            "getSession: Returning TRY_REFRESH_TOKEN because failed to decode access token"
        )
        raise_try_refresh_token_exception(e)


def validate_access_token_structure(payload: Dict[str, Any]) -> None:
    if (
        not isinstance(payload.get("sessionHandle"), str)
        or payload.get("userData") is None
        or not isinstance(payload.get("refreshTokenHash1"), str)
        or not isinstance(payload.get("expiryTime"), int)
        or not isinstance(payload.get("timeCreated"), int)
    ):
        raise Exception(
            "Access token does not contain all the information. Maybe the structure has changed?"
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Dict, Union

if TYPE_CHECKING:
    from .types import AccessTokenAPI, AuthorisationRedirectAPI, UserInfo


class Provider(abc.ABC):
    def __init__(self, provider_id: str, is_default: bool):
        self.id: str = provider_id
        self.is_default: bool = is_default

    @abc.abstractmethod
    async def get_profile_info(
        self, auth_code_response: Dict[str, Any], user_context: Dict[str, Any]
    ) -> UserInfo:
        pass

    @abc.abstractmethod
    def get_authorisation_redirect_api_info(
        self, user_context: Dict[str, Any]
    ) -> AuthorisationRedirectAPI:
        pass

    @abc.abstractmethod
    def get_access_token_api_info(
        self,
        redirect_uri: str,
        auth_code_from_request: str,
        user_context: Dict[str, Any],
    ) -> AccessTokenAPI:
        pass

    @abc.abstractmethod
    def get_redirect_uri(self, user_context: Dict[str, Any]) -> Union[None, str]:
        pass

    @abc.abstractmethod
    def get_client_id(self, user_context: Dict[str, Any]) -> str:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Set, Union

from supertokens_python.exceptions import raise_bad_input_exception

from .interfaces import APIInterface, RecipeInterface

if TYPE_CHECKING:
    from .provider import Provider

from jwt import PyJWKClient, decode


class SignInAndUpFeature:
    def __init__(self, providers: List[Provider]):
        if len(providers) == 0:
            raise_bad_input_exception(
                "thirdparty recipe requires atleast 1 provider to be passed in sign_in_and_up_feature.providers config"
            )
        default_providers_set: Set[str] = set()
        all_providers_set: Set[str] = set()

        for provider in providers:
            provider_id = provider.id
            all_providers_set.add(provider_id)
            is_default = provider.is_default

            if not is_default:
                # if this id is not being used by any other provider, we treat
                # this as the is_default
                other_providers_with_same_id = list(
                    filter(lambda p: p.id == provider_id and provider != p, providers)
                )
                if len(other_providers_with_same_id) == 0:
                    # we treat this as the isDefault now.
                    is_default = True
            if is_default:
                if provider_id in default_providers_set:
                    raise_bad_input_exception(
                        'You have provided multiple third party providers that have the id: "'
                        + provider_id
                        + '" '
                        "and "
                        "are "
                        'marked as "is_default: True". Please only mark one of them as is_default.'
                    )
                default_providers_set.add(provider_id)

        if len(default_providers_set) != len(all_providers_set):
            # this means that there is no provider marked as is_default
            raise_bad_input_exception(
                "The providers array has multiple entries for the same third party provider. Please "
                'mark one of them as the default one by using "is_default: true".'
            )
        self.providers = providers


class InputOverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class ThirdPartyConfig:
    def __init__(
        self,
        sign_in_and_up_feature: SignInAndUpFeature,
        override: OverrideConfig,
    ):
        self.sign_in_and_up_feature = sign_in_and_up_feature
        self.override = override


def validate_and_normalise_user_input(
    sign_in_and_up_feature: SignInAndUpFeature,
    override: Union[InputOverrideConfig, None] = None,
) -> ThirdPartyConfig:
    if not isinstance(sign_in_and_up_feature, SignInAndUpFeature):  # type: ignore
        raise ValueError(
            "sign_in_and_up_feature must be an instance of SignInAndUpFeature"
        )

    if override is not None and not isinstance(override, InputOverrideConfig):  # type: ignore
        raise ValueError("override must be an instance of InputOverrideConfig or None")

    if override is None:
        override = InputOverrideConfig()

    return ThirdPartyConfig(
        sign_in_and_up_feature,
        OverrideConfig(functions=override.functions, apis=override.apis),
    )


def find_right_provider(
    providers: List[Provider], third_party_id: str, client_id: Union[str, None]
) -> Union[Provider, None]:
    for provider in providers:
        provider_id = provider.id
        if provider_id != third_party_id:
            continue

        # first if there is only one provider with third_party_id in the
        # providers array
        other_providers_with_same_id = list(
            filter(lambda p: p.id == provider_id and provider != p, providers)
        )
        if len(other_providers_with_same_id) == 0:
            # then we always return that.
            return provider

        # otherwise, we look for the is_default provider if client_id is
        # missing
        if client_id is None and provider.is_default:
            return provider

        # otherwise, we return a provider that matches based on client Id as
        # well.
        if provider.get_client_id({}) == client_id:
            return provider

    return None


def verify_id_token_from_jwks_endpoint(
    id_token: str, jwks_uri: str, audience: str, issuers: List[str]
) -> Dict[str, Any]:
    jwks_client = PyJWKClient(jwks_uri)
    signing_key = jwks_client.get_signing_key_from_jwt(id_token)

    data: Dict[str, Any] = decode(  # type: ignore
        id_token,
        signing_key.key,  # type: ignore
        algorithms=["RS256"],
        audience=audience,
        options={"verify_exp": False},
    )

    issuer_found = False
    for issuer in issuers:
        if data["iss"] == issuer:
            issuer_found = True

    if not issuer_found:
        raise Exception("no required issuer found")

    return data



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, Union

from supertokens_python.recipe.jwt.interfaces import (
    CreateJwtOkResult,
    CreateJwtResultUnsupportedAlgorithm,
    GetJWKSResult,
)
from supertokens_python.recipe.jwt.recipe import JWTRecipe


async def create_jwt(
    payload: Union[None, Dict[str, Any]] = None,
    validity_seconds: Union[None, int] = None,
    user_context: Union[Dict[str, Any], None] = None,
) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
    if user_context is None:
        user_context = {}
    if payload is None:
        payload = {}

    return await JWTRecipe.get_instance().recipe_implementation.create_jwt(
        payload, validity_seconds, user_context
    )


async def get_jwks(user_context: Union[Dict[str, Any], None] = None) -> GetJWKSResult:
    if user_context is None:
        user_context = {}
    return await JWTRecipe.get_instance().recipe_implementation.get_jwks(user_context)
